import * as PIXI from "pixi.js";
import { TextureMatrix } from "pixi.js";
import seedrandom from "seedrandom";
import { genButton } from "../button.js";

import globals, { store } from "../globals.js";
import { randomFrom } from "../helper.js";
import { setStage } from "../stages.js";

import { afterPuzzle, Stage } from "./stage.js";

const stage = new Stage("cable_puzzle", setup, activate);
export default stage;

const colors = {
    red: 0xff0000,
    green: 0x00ff00,
    blue: 0x0000ff,
    yellow: 0xffff00,
    purple: 0xff00ff,
    turquoise: 0x00ffff,
    orange: 0xff8000,
    white: 0xffffff,
};

let stageWidth = 0.8;

let graphics;

let dragObj = {
    line: null,
    srcSocket: null,
    srcLayer: null,
    isDragging: false,
};

let ordering;
let routes;
let visibleRoutes;
let missingRoutes;

let manualRoutes;
let manualConnectionContainer;

let nextStageButton;
let routesFoundText;

function onDown(socket, layerIndex) {
    return () => {
        for (const route of visibleRoutes) {
            if (
                (route.from.layerIndex === layerIndex &&
                    route.from.socket === socket) ||
                (route.to.layerIndex == layerIndex &&
                    route.to.socket === socket)
            ) {
                return;
            }
        }

        if (dragObj.isDragging && layerIndex === dragObj.srcLayer) {
            return;
        }

        manualRoutes = manualRoutes.filter((route) => {
            return !(
                (route.from.layerIndex === layerIndex &&
                    route.from.socket === socket) ||
                (route.to.layerIndex === layerIndex &&
                    route.to.socket === socket)
            );
        });

        if (!dragObj.isDragging) {
            dragObj.srcSocket = socket;
            dragObj.srcLayer = layerIndex;
            dragObj.isDragging = true;
        } else {
            dragObj.isDragging = false;
            dragObj.line.clear();
            const from =
                dragObj.srcLayer < layerIndex
                    ? {
                          layerIndex: dragObj.srcLayer,
                          socket: dragObj.srcSocket,
                      }
                    : {
                          layerIndex,
                          socket,
                      };
            const to =
                dragObj.srcLayer < layerIndex
                    ? {
                          layerIndex,
                          socket,
                      }
                    : {
                          layerIndex: dragObj.srcLayer,
                          socket: dragObj.srcSocket,
                      };

            manualRoutes.push({ from, to });
            manualConnectionContainer.removeChildren();
            for (const route of manualRoutes) {
                const connection = drawConnection(route);
                manualConnectionContainer.addChild(connection);
            }
        }
        checkWinCondition();
    };
}

function updateDragLine(event) {
    if (dragObj.line === null || !dragObj.isDragging) {
        return;
    }

    dragObj.line.clear();
    const color =
        dragObj.srcLayer === 1
            ? 0xaaaaaa
            : colors[ordering[dragObj.srcLayer][dragObj.srcSocket]];
    dragObj.line.lineStyle({
        width: 8,
        color: color,
        cap: PIXI.LINE_CAP.ROUND,
    });

    const isUpper =
        (store.player === "A" && dragObj.srcLayer === 0) ||
        (store.player === "B" && dragObj.srcLayer === 1);
    const startX =
        (dragObj.srcSocket * window.innerWidth * stageWidth) /
            (ordering[dragObj.srcLayer].length - 1) +
        (window.innerWidth * (1 - stageWidth)) / 2;

    const startY = (window.innerHeight * (isUpper ? 1 : 3)) / 4;

    dragObj.line.moveTo(startX, startY);
    dragObj.line.lineTo(event.data.global.x, event.data.global.y);
}

function checkWinCondition() {
    let routesFound = 0;
    for (const missingRoute of missingRoutes) {
        for (const manualRoute of manualRoutes) {
            if (
                manualRoute.from.layerIndex === missingRoute.from.layerIndex &&
                manualRoute.from.socket === missingRoute.from.socket &&
                manualRoute.to.layerIndex === missingRoute.to.layerIndex &&
                manualRoute.to.socket === missingRoute.to.socket
            ) {
                routesFound += 1;
            }
        }
    }

    nextStageButton.visible = routesFound == missingRoutes.length;
    routesFoundText.text = `${routesFound} / 3`;
}

function drawArrow() {
    const arrow = new PIXI.Graphics();

    const heightProportion = 0.6;
    const widthProportion = 0.5;

    const arrowHeight = window.innerHeight * 0.2;
    const arrowWidth = arrowHeight * 0.3;

    arrow.beginFill(0xffffff);
    arrow.drawRect(
        arrowWidth * ((1 - widthProportion) / 2),
        0,
        arrowWidth * widthProportion,
        arrowHeight * heightProportion
    );

    arrow.drawPolygon([
        new PIXI.Point(0, arrowHeight * heightProportion),
        new PIXI.Point(arrowWidth, arrowHeight * heightProportion),
        new PIXI.Point(arrowWidth / 2, arrowHeight),
    ]);

    arrow.endFill();
    return arrow;
}

function drawLayer(layerIndex) {
    const layer = new PIXI.Container();

    for (const [i, color] of Object.entries(ordering[layerIndex])) {
        const socket = new PIXI.Graphics();

        const hasFixedRoute =
            visibleRoutes.filter(
                (route) =>
                    (route.from.layerIndex === layerIndex &&
                        route.from.socket.toString() === i) ||
                    (route.to.layerIndex === layerIndex &&
                        route.to.socket.toString() === i)
            ).length > 0;

        const colorValue = layerIndex === 1 ? 0xaaaaaa : colors[color];
        if (hasFixedRoute) {
            socket.beginFill(colorValue);
        } else {
            socket.lineStyle(2, colorValue);
        }

        const r = window.innerHeight * 0.02;
        const x =
            (i * window.innerWidth * stageWidth) /
            (ordering[layerIndex].length - 1);

        socket.drawCircle(x, 0, r);
        if (hasFixedRoute) {
            socket.endFill();
        }

        socket.interactive = true;
        socket.buttonMode = true;
        socket.hitArea = new PIXI.Polygon([
            new PIXI.Point(x - r, -r),
            new PIXI.Point(x + r, -r),
            new PIXI.Point(x + r, r),
            new PIXI.Point(x - r, r),
        ]);
        socket.on("pointerdown", onDown(parseInt(i), layerIndex));
        layer.addChild(socket);
    }

    return layer;
}

function drawConnection(route) {
    const connection = new PIXI.Graphics();

    const upperX =
        (route.from.socket * window.innerWidth * stageWidth) /
            (ordering[route.from.layerIndex].length - 1) +
        (window.innerWidth * (1 - stageWidth)) / 2;
    const lowerX =
        (route.to.socket * window.innerWidth * stageWidth) /
            (ordering[route.to.layerIndex].length - 1) +
        (window.innerWidth * (1 - stageWidth)) / 2;

    const color =
        colors[
            store.player === "A"
                ? ordering[route.from.layerIndex][route.from.socket]
                : ordering[route.to.layerIndex][route.to.socket]
        ];

    connection.lineStyle({
        width: 8,
        color: color,
        cap: PIXI.LINE_CAP.ROUND,
    });
    connection.moveTo(upperX, window.innerHeight / 4);
    connection.lineTo(lowerX, (window.innerHeight * 3) / 4);

    return connection;
}

function drawGraphics() {
    const topLayerIndex = store.player === "A" ? 0 : 1;
    const bottomLayerIndex = store.player === "A" ? 1 : 2;

    const topLayer = drawLayer(topLayerIndex);
    topLayer.position.set(
        window.innerWidth * ((1 - stageWidth) / 2),
        window.innerHeight / 4
    );

    const bottomLayer = drawLayer(bottomLayerIndex);
    bottomLayer.position.set(
        window.innerWidth * ((1 - stageWidth) / 2),
        (window.innerHeight * 3) / 4
    );

    for (const route of visibleRoutes) {
        const connnection = drawConnection(route);
        graphics.addChild(connnection);
    }

    const arrow = drawArrow();
    const arrowBounds = arrow.getBounds();
    arrow.position.set(
        window.innerWidth * ((1 - stageWidth) / 4) - arrowBounds.width / 2,
        window.innerHeight / 2 - arrowBounds.height / 2
    );

    graphics.addChild(arrow);
    graphics.addChild(topLayer);
    graphics.addChild(bottomLayer);

    return graphics;
}

function genData() {
    ordering = Array(3);
    routes = [];

    for (let layerNum = 0; layerNum < ordering.length; layerNum++) {
        ordering[layerNum] = Array(Object.keys(colors).length);
        for (
            let socketNum = 0;
            socketNum < ordering[layerNum].length;
            socketNum++
        ) {
            ordering[layerNum][socketNum] = randomFrom(
                rand,
                Object.keys(colors).filter((col) => {
                    for (const color of ordering[layerNum]) {
                        if (color === col) {
                            return false;
                        }
                    }
                    return true;
                })
            );
        }
    }

    const usedRouteColors = [];
    const usedViaColors = [];
    for (let i = 0; i < 6; i++) {
        const viaColor = randomFrom(
            rand,
            Object.keys(colors).filter((col) => {
                for (const used of usedViaColors) {
                    if (used === col) {
                        return false;
                    }
                }
                return true;
            })
        );
        usedViaColors.push(viaColor);
        const routeColor = randomFrom(
            rand,
            Object.keys(colors).filter((col) => {
                for (const used of usedRouteColors) {
                    if (used === col) {
                        return false;
                    }
                }
                return true;
            })
        );
        usedRouteColors.push(routeColor);

        routes.push({
            from: {
                layerIndex: 0,
                socket: ordering[0].indexOf(routeColor),
            },
            to: {
                layerIndex: 1,
                socket: ordering[1].indexOf(viaColor),
            },
        });
        routes.push({
            from: {
                layerIndex: 1,
                socket: ordering[1].indexOf(viaColor),
            },
            to: {
                layerIndex: 2,
                socket: ordering[2].indexOf(routeColor),
            },
        });
    }

    const topLayerIndex = store.player === "A" ? 0 : 1;
    const bottomLayerIndex = store.player === "A" ? 1 : 2;

    visibleRoutes = routes
        .filter(
            (route) =>
                route.from.layerIndex === topLayerIndex ||
                route.to.layerIndex == bottomLayerIndex
        )
        .filter((_, n) => n % 2 === (store.player === "A" ? 0 : 1));
    missingRoutes = routes
        .filter(
            (route) =>
                route.from.layerIndex === topLayerIndex ||
                route.to.layerIndex == bottomLayerIndex
        )
        .filter((_, n) => n % 2 === (store.player === "A" ? 1 : 0));
}

function setup() {
    dragObj.line = new PIXI.Graphics();

    nextStageButton = genButton("Next Stage", afterPuzzle);
    nextStageButton.position.set(
        window.innerWidth / 2,
        (window.innerHeight * 7) / 8
    );
    nextStageButton.visible = false;

    routesFoundText = new PIXI.Text("0 / 0", globals.defaultStyle);
    routesFoundText.anchor.set(0.5, 0.5);
    routesFoundText.position.set(window.innerWidth / 2, window.innerHeight / 8);

    manualConnectionContainer = new PIXI.Container();

    graphics = new PIXI.Container();

    stage.addChild(dragObj.line);
    stage.addChild(nextStageButton);
    stage.addChild(routesFoundText);
    stage.addChild(manualConnectionContainer);
    stage.addChild(graphics);
}

let rand;
function activate() {
    rand = new seedrandom(store.code + store.finishedPuzzles.toString());

    manualRoutes = [];
    manualConnectionContainer.removeChildren();
    graphics.removeChildren();
    genData();
    drawGraphics();

    stage.interactive = true;
    stage.hitArea = new PIXI.Polygon([
        new PIXI.Point(0, 0),
        new PIXI.Point(window.innerWidth, 0),
        new PIXI.Point(window.innerWidth, window.innerHeight),
        new PIXI.Point(0, window.innerHeight),
    ]);
    stage.on("pointermove", updateDragLine);
    checkWinCondition();
}
