import * as PIXI from "pixi.js";
import { TextureMatrix } from "pixi.js";
import seedrandom from "seedrandom";
import { genButton } from "../button.js";

import globals, { store } from "../globals.js";
import { randomFrom } from "../helper.js";
import { setStage } from "../stages.js";

import { afterPuzzle, Stage } from "./stage.js";

const stage = new Stage("word_puzzle", setup, activate);
export default stage;

const words = [
    "VIRTUE",
    "OPTION",
    "STITCH",
    "NORMAL",
    "SENIOR",
    "WINDOW",
    "LAWYER",
    "VIABLE",
    "ARTIST",
    "THROAT",
    "MATURE",
    "HUNTER",
    "DEFINE",
    "EXTORT",
    "GARAGE",
    "CATTLE",
    "COUSIN",
    "UNLIKE",
    "RESIGN",
    "ACTION",
    "TICKET",
    "JUNGLE",
    "SHIVER",
];
const letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

let stageWidth = 0.8;
let nextStageButton;
let wordsList;

let solution;
let lettersSet;
let rows;
let rowObjects;

function checkWinCondition() {
    let needed;
    let entered;
    if (store.player === "A") {
        needed = solution.substr(0, solution.length / 2);
        entered = lettersSet.substr(0, lettersSet.length / 2);
    } else if (store.player === "B") {
        needed = solution.substr(solution.length / 2);
        entered = lettersSet.substr(lettersSet.length / 2);
    }
    nextStageButton.visible = needed === entered;
}

function makeTriangleButton(dir, size) {
    size *= 0.5;
    const height = Math.sqrt(size ** 2 - (size / 2) ** 2);

    const triangle = new PIXI.Graphics();
    triangle.beginFill(0xffffff);
    triangle.drawRegularPolygon(
        0,
        height,
        size,
        3,
        Math.PI * (dir === "up" ? 0 : 1)
    );
    triangle.endFill();

    return triangle;
}

function makeRow(handler, rowIndex) {
    const handle = (dir) => {
        return () => {
            const offset = dir === "up" ? 1 : -1;
            let newPos =
                (rows[rowIndex].indexOf(letter.text) + offset) %
                rows[rowIndex].length;
            // Bug in JS Modulo
            if (newPos < 0) {
                newPos += rows[rowIndex].length;
            }
            letter.text = rows[rowIndex][newPos];
            handler(dir, rows[rowIndex][newPos]);
        };
    };

    const row = new PIXI.Container();
    row.customHandler = handle();

    const letter = new PIXI.Text(lettersSet[rowIndex], {
        ...globals.defaultStyle,
        fontSize: globals.defaultStyle.fontSize * 1.5,
    });
    let letterBounds = letter.getBounds();
    letter.anchor.set(0.5, 0);
    letter.position.y = letterBounds.height;
    letter.position.x = 0;

    const upButton = makeTriangleButton("up", letterBounds.height);
    upButton.position.set(0, 0);
    upButton.interactive = true;
    upButton.buttonMode = true;
    upButton.on("pointerdown", handle("up"));

    const downButton = makeTriangleButton("down", letterBounds.height);
    downButton.position.set(0, letterBounds.height * 2);
    downButton.interactive = true;
    downButton.buttonMode = true;
    downButton.on("pointerdown", handle("down"));

    row.addChild(upButton);
    row.addChild(letter);
    row.addChild(downButton);

    const bounds = row.getBounds();
    row.position.set(
        window.innerWidth / 2,
        (window.innerHeight - bounds.height) / 2
    );

    return row;
}

function rowHandler(pos) {
    return (dir, letter) => {
        let firstPart = lettersSet.substr(0, pos);
        let lastPart = lettersSet.substr(pos + 1);

        lettersSet = firstPart + letter + lastPart;
        checkWinCondition();
    };
}

function genRows() {
    const inputWidth = window.innerWidth * stageWidth;
    for (const [i, rowData] of Object.entries(rows)) {
        const row = makeRow(rowHandler(parseInt(i)), parseInt(i));
        row.position.x +=
            -inputWidth / 2 + parseInt(i) * (inputWidth / (rows.length - 1));
        stage.addChild(row);
        rowObjects.push(row);
    }
}

function genData() {
    solution = randomFrom(rand, words);

    const rowSize = 5;

    let possibleWords = words;

    rows = Array(solution.length);

    const maxTries = 100;
    let tryNo = 0;

    lettersSet = "";
    while (possibleWords.length !== 1) {
        if (tryNo >= maxTries) {
            setStage("ohno");
            return;
        }

        for (let i = 0; i < rows.length; i++) {
            rows[i] = [solution[i]];
            for (let j = 0; j < rowSize - 1; j++) {
                rows[i].push(
                    randomFrom(
                        rand,
                        letters
                            .split("")
                            .filter((letter) => !rows[i].includes(letter))
                    )
                );
            }

            const remainingWords = [];
            for (const [wordIndex, word] of Object.entries(possibleWords)) {
                for (const letter of rows[i]) {
                    if (word[i] === letter) {
                        remainingWords.push(word);
                    }
                }
            }
            possibleWords = remainingWords;
            lettersSet += randomFrom(rand, rows[i]);
        }
        tryNo += 1;
    }
}

function setup() {
    rowObjects = [];

    nextStageButton = genButton("Next Stage", afterPuzzle);
    nextStageButton.position.set(
        window.innerWidth / 2,
        (window.innerHeight * 7) / 8
    );
    nextStageButton.visible = false;

    lettersSet = "A".repeat(words[0].length);
    rows = Array.from(Array(words[0].length).keys()).map(() => ["A"]);
    genRows();

    wordsList = new PIXI.Text("", globals.defaultStyle);
    wordsList.anchor.set(0.5, 0.5);
    wordsList.position.set(window.innerWidth / 2, window.innerHeight / 4);

    stage.addChild(wordsList);
    stage.addChild(nextStageButton);
}

let rand;
function activate() {
    rand = new seedrandom(store.code + store.finishedPuzzles.toString());

    wordsList.text = words
        .filter((_, i) => i % 2 === (store.player === "A" ? 0 : 1))
        .join(", ");

    genData();
    for (const [i, row] of Object.entries(rowObjects)) {
        if (
            (store.player === "A" && parseInt(i) >= rows.length / 2) ||
            (store.player === "B" && parseInt(i) < rows.length / 2)
        ) {
            row.visible = false;
        } else {
            row.visible = true;
            row.customHandler("up");
        }
    }

    checkWinCondition();
}
